import math
from typing import Any, Callable, Dict, List, Tuple, TypeVar

from grug import GrugPackage
from grug.entity import GameFnError

try:
    from typing import Protocol  # Python >= 3.8
except ImportError:
    from typing_extensions import Protocol  # Python 3.7


# --------------------
# Assertions
# --------------------


def assert_bool(b1: bool, b2: bool):
    assert b1 == b2, f"assert_bool failed: {b1} != {b2}"


def assert_id(id1: object, id2: object):
    assert id1 == id2, f"assert_id failed: {id1} != {id2}"


def assert_number(n1: float, n2: float):
    assert n1 == n2, f"assert_number failed: {n1} != {n2}"


def assert_string(s1: str, s2: str):
    assert s1 == s2, f"assert_string failed: '{s1}' != '{s2}'"


# --------------------
# Math
# --------------------


def ceil(n: float) -> float:
    return float(math.ceil(n))


def sqrt(n: float) -> float:
    return math.sqrt(n)


# --------------------
# Dict core
# --------------------


def id_to_dict(id_: Dict[object, object]) -> Dict[object, object]:
    return id_


def dict_len(d: Dict[object, object]) -> float:
    return float(len(d))


def dict_X() -> Dict[object, object]:
    return {}


dict_X.__name__ = "dict"


def dict_set(d: Dict[object, object], key: object, val: object):
    d[key] = val


def dict_has_key(d: Dict[object, object], key: object) -> bool:
    return key in d


def dict_get(d: Dict[object, object], key: object) -> object:
    value = d.get(key)
    if value is None:
        raise GameFnError(
            f"dict_get({d}, {key}) failed, as key '{key}' is not in the Dict"
        )
    return value


def dict_get_default(d: Dict[object, object], key: object, default: object) -> object:
    return d.get(key, default)


def dict_set_default(d: Dict[object, object], key: object, val: object) -> object:
    return d.setdefault(key, val)


def dict_pop(d: Dict[object, object], key: object) -> object:
    return d.pop(key)


def dict_update(d: Dict[object, object], other: Dict[object, object]):
    d.update(other)


def dict_fromkeys(keys: List[object], val: object) -> Dict[object, object]:
    return dict.fromkeys(keys, val)


def dict_copy(d: Dict[object, object]) -> Dict[object, object]:
    return d.copy()


def dict_clear(d: Dict[object, object]):
    d.clear()


def dict_keys(d: Dict[object, object]) -> List[object]:
    return list(d.keys())


def dict_values(d: Dict[object, object]) -> List[object]:
    return list(d.values())


def dict_items(d: Dict[object, object]) -> List[List[object]]:
    return [[k, v] for k, v in d.items()]


def dict_popitem(d: Dict[object, object]) -> List[object]:
    k, v = d.popitem()
    return [k, v]


# --------------------
# List core
# --------------------


def id_to_list(id_: List[object]) -> List[object]:
    return id_


def list_clear(l: List[object]):
    l.clear()


def list_copy(l: List[object]) -> List[object]:
    return l.copy()


def list_extend(lst1: List[object], lst2: List[object]):
    lst1.extend(lst2)


def list_len(l: List[object]) -> float:
    return float(len(l))


def list_reverse(l: List[object]):
    l.reverse()


class SupportsLessThan(Protocol):
    def __lt__(self, __other: object) -> bool: ...


T = TypeVar("T", bound=SupportsLessThan)


def list_sort(l: List[T]):
    l.sort()


def list_X() -> List[object]:
    return []


list_X.__name__ = "list"


def list_append(l: List[object], val: object):
    l.append(val)


def list_count(l: List[object], val: object) -> float:
    return float(l.count(val))


def list_index(l: List[object], val: object) -> float:
    return float(l.index(val))


def list_insert(l: List[object], index: float, val: object):
    l.insert(int(index), val)


def list_pop(l: List[object]):
    return l.pop()


def list_pop_index(l: List[object], index: float):
    return l.pop(int(index))


def list_remove(l: List[object], val: object):
    l.remove(val)


# --------------------
# Printing
# --------------------


def print_bool(b: bool):
    print(b)


def print_id(id: object):
    print(id)


def format_number(x: object) -> object:
    if isinstance(x, float) and x.is_integer():
        return int(x)
    return x


def print_list(l: List[object]):
    print([format_number(x) for x in l])


def print_dict(d: Dict[object, object]):
    print({format_number(k): format_number(v) for k, v in d.items()})


def print_number(n: float):
    print(int(n) if n.is_integer() else n)


def print_string(s: str):
    print(s)


# --------------------
# Game fn registration
# --------------------


def assert_fns() -> List[Callable[..., Any]]:
    return [
        assert_bool,
        assert_id,
        assert_number,
        assert_string,
    ]


def casting_fns() -> List[Callable[..., Any]]:
    return [
        id_to_dict,
        id_to_list,
    ]


def math_fns() -> List[Callable[..., Any]]:
    return [
        ceil,
        sqrt,
    ]


def printing_fns() -> List[Callable[..., Any]]:
    return [
        print_bool,
        print_id,
        print_list,
        print_dict,
        print_number,
        print_string,
    ]


# --------------------
# Container registration
# --------------------


def wrap(fn: Callable[..., Any], name: str) -> Callable[..., Any]:
    def wrapper(*args: Tuple[Any, ...], **kwargs: Dict[str, Any]) -> Any:
        return fn(*args, **kwargs)

    wrapper.__name__ = name
    return wrapper


HASHABLE_TYPES = ("number", "bool", "string")
VALUE_TYPES = ("number", "bool", "string", "id")


def dict_fns() -> List[Callable[..., Any]]:
    fns: List[Callable[..., Any]] = []

    for fn in [
        dict_X,
        dict_len,
        dict_keys,
        dict_values,
        dict_items,
        dict_popitem,
        dict_update,
        dict_copy,
        dict_clear,
    ]:
        fns.append(fn)

    for value_type in VALUE_TYPES:
        fns.append(wrap(dict_pop, f"dict_{value_type}_pop"))
        fns.append(wrap(dict_fromkeys, f"dict_{value_type}_fromkeys"))

    for key_type in HASHABLE_TYPES:
        fns.append(wrap(dict_has_key, f"dict_{key_type}_has_key"))

        for value_type in VALUE_TYPES:
            fns.append(wrap(dict_get, f"dict_{key_type}_{value_type}_get"))
            fns.append(
                wrap(dict_get_default, f"dict_{key_type}_{value_type}_get_default")
            )
            fns.append(wrap(dict_set, f"dict_{key_type}_{value_type}_set"))
            fns.append(
                wrap(dict_set_default, f"dict_{key_type}_{value_type}_set_default")
            )

    return fns


def list_fns() -> List[Callable[..., Any]]:
    fns: List[Callable[..., Any]] = []

    for fn in [
        list_X,
        list_len,
        list_sort,
        list_clear,
        list_copy,
        list_extend,
        list_reverse,
    ]:
        fns.append(fn)

    for value_type in VALUE_TYPES:
        fns.append(wrap(list_append, f"list_{value_type}_append"))
        fns.append(wrap(list_count, f"list_{value_type}_count"))
        fns.append(wrap(list_index, f"list_{value_type}_index"))
        fns.append(wrap(list_insert, f"list_{value_type}_insert"))
        fns.append(wrap(list_pop, f"list_{value_type}_pop"))
        fns.append(wrap(list_pop_index, f"list_{value_type}_pop_index"))
        fns.append(wrap(list_remove, f"list_{value_type}_remove"))

    return fns


# --------------------
# Package
# --------------------


def get():
    game_fns: List[Callable[..., Any]] = []

    game_fns.extend(assert_fns())
    game_fns.extend(casting_fns())
    game_fns.extend(math_fns())
    game_fns.extend(printing_fns())

    # Containers
    game_fns.extend(dict_fns())
    game_fns.extend(list_fns())

    return GrugPackage(
        prefix="",
        game_fns=game_fns,
    )



import ctypes
import sys
import traceback
from pathlib import Path
from typing import List, Optional, Union

import pytest

import grug
from grug.entity import Entity, ReraisedGameFnError, StackOverflow, TimeLimitExceeded
from grug.grug_file import GrugFile
from grug.grug_state import GrugRuntimeErrorType, GrugState
from grug.grug_value import GrugValue


class GrugValueUnion(ctypes.Union):
    _fields_ = [
        ("_number", ctypes.c_double),
        ("_bool", ctypes.c_bool),
        ("_string", ctypes.c_char_p),
        ("_id", ctypes.c_uint64),
    ]


class GrugValueWorkaround(ctypes.Structure):
    """
    This defines a Structure of the exact same size and alignment as GrugValueUnion.

    When using `ctypes.Union`, Python's logic for "return by value" is flawed.
    It seemingly assumes complex types (like Unions) are always too large
    for registers and must be returned via memory. It allocates a buffer,
    passes its address to C (which C ignores), and then reads that buffer back.
    Since C never wrote to it, you see garbage memory.

    Quoting a [cpython GitHub issue](https://github.com/python/cpython/issues/60779) from 2012:
    > ctypes pretends to support passing arguments to C functions
    > that are unions (not pointers to unions), but that's a lie.
    > In fact, the underlying libffi does not support it.
    """

    _fields_ = [("_blob", ctypes.c_uint64)]


_g_grug_lib: ctypes.PyDLL

_grug_runtime_err: Optional[
    Union[TimeLimitExceeded, StackOverflow, ReraisedGameFnError]
] = None


def custom_runtime_error_handler(
    reason: str,
    grug_runtime_error_type: GrugRuntimeErrorType,
    on_fn_name: str,
    on_fn_path: str,
):
    _g_grug_lib.grug_tests_runtime_error_handler(
        reason.encode(),
        grug_runtime_error_type.value,
        on_fn_name.encode(),
        on_fn_path.encode(),
    )


def c_to_py_value(value: GrugValueUnion, typ: str):
    if typ == "number":
        return float(value._number)
    if typ == "bool":
        return bool(value._bool)
    if typ == "string":
        return ctypes.string_at(value._string).decode()
    return int(value._id)


def test_grug(
    grug_tests_path: Path, whitelisted_test: Optional[str], grug_lib: ctypes.PyDLL
) -> None:
    global _g_grug_lib
    _g_grug_lib = grug_lib

    state = grug.init(
        runtime_error_handler=custom_runtime_error_handler,
        mod_api_path=str(grug_tests_path / "mod_api.json"),
        mods_dir_path=str(grug_tests_path / "tests"),
    )

    GameFnRegistrator(state, grug_lib).register_game_fns()

    grug_file: Optional[GrugFile] = None
    grug_entity: Optional[Entity] = None

    @ctypes.CFUNCTYPE(ctypes.c_char_p, ctypes.c_char_p)
    def compile_grug_file(path: bytes) -> Optional[bytes]:
        nonlocal grug_file
        try:
            grug_file = state.compile_grug_file(path.decode())
        except Exception as e:
            return str(e).encode()
        return None

    @ctypes.CFUNCTYPE(None)
    def init_globals_fn_dispatcher() -> None:
        try:
            global _grug_runtime_err
            _grug_runtime_err = None

            state.next_id = 42

            nonlocal grug_entity
            assert grug_file is not None
            grug_entity = grug_file.create_entity()
        except (TimeLimitExceeded, StackOverflow, ReraisedGameFnError) as e:
            # Necessary, as propagating exceptions from
            # this CFUNCTYPE function doesn't work.
            _grug_runtime_err = e
        except Exception:
            traceback.print_exc(file=sys.stderr)

    on_fn_name: Optional[str] = None

    @ctypes.CFUNCTYPE(None, ctypes.c_char_p, ctypes.POINTER(GrugValueUnion))
    def on_fn_dispatcher(c_on_fn_name: bytes, c_args: List[GrugValueUnion]) -> None:
        try:
            global _grug_runtime_err
            _grug_runtime_err = None

            nonlocal on_fn_name
            on_fn_name = c_on_fn_name.decode()

            assert grug_file
            on_fn_decl = grug_file.on_fns.get(on_fn_name)
            if not on_fn_decl:
                raise RuntimeError(
                    f"The function '{on_fn_name}' is not defined by the file {grug_file.relative_path}"
                )

            args = [
                c_to_py_value(arg, argument.type_name)
                for arg, argument in zip(c_args or [], on_fn_decl.arguments)
            ]

            assert grug_entity is not None
            on_fn = getattr(grug_entity, on_fn_name)
            on_fn(*args)
        except (TimeLimitExceeded, StackOverflow, ReraisedGameFnError) as e:
            # Necessary, as propagating exceptions from
            # this CFUNCTYPE function doesn't work.
            _grug_runtime_err = e
        except Exception:
            traceback.print_exc(file=sys.stderr)

    @ctypes.CFUNCTYPE(ctypes.c_bool, ctypes.c_char_p, ctypes.c_char_p)
    def dump_file_to_json(input_grug_path: bytes, output_json_path: bytes) -> bool:
        return state.dump_file_to_json(
            input_grug_path.decode(), output_json_path.decode()
        )

    @ctypes.CFUNCTYPE(ctypes.c_bool, ctypes.c_char_p, ctypes.c_char_p)
    def generate_file_from_json(
        input_json_path: bytes, output_grug_path: bytes
    ) -> bool:
        return state.generate_file_from_json(
            input_json_path.decode(), output_grug_path.decode()
        )

    _original_run_game_fn = Entity._run_game_fn  # pyright: ignore[reportPrivateUsage]

    _game_fn_error_reason: Optional[str] = None

    def _test_run_game_fn(
        self: Entity, name: str, *args: GrugValue
    ) -> Optional[GrugValue]:
        nonlocal _game_fn_error_reason

        # Call the original method
        result = _original_run_game_fn(self, name, *args)

        # Raise _game_fn_error_reason if it's not None
        if _game_fn_error_reason is not None:
            reason = _game_fn_error_reason
            self.state.runtime_error_handler(
                reason,
                GrugRuntimeErrorType.GAME_FN_ERROR,
                self.fn_name,
                self.file.relative_path,
            )

            _game_fn_error_reason = None
            raise ReraisedGameFnError(reason)

        return result

    # Patch the method for testing
    Entity._run_game_fn = _test_run_game_fn  # pyright: ignore[reportPrivateUsage]

    @ctypes.CFUNCTYPE(None, ctypes.c_char_p)
    def game_fn_error(reason: bytes) -> None:
        nonlocal _game_fn_error_reason
        _game_fn_error_reason = ctypes.string_at(reason).decode()

    print("\n")

    grug_lib.grug_tests_run(
        str(grug_tests_path / "tests").encode(),
        compile_grug_file,
        init_globals_fn_dispatcher,
        on_fn_dispatcher,
        dump_file_to_json,
        generate_file_from_json,
        game_fn_error,
        whitelisted_test.encode() if whitelisted_test else None,
    )


class GameFnRegistrator:
    def __init__(self, state: GrugState, grug_lib: ctypes.PyDLL):
        self.state = state
        self.grug_lib = grug_lib

    def register_game_fns(self):
        self._register_void_argless("nothing")
        self._register_value_argless("magic")
        self._register_void("initialize")
        self._register_void("initialize_bool")
        self._register_value("identity")
        self._register_value("max")
        self._register_void("say")
        self._register_value("sin")
        self._register_value("cos")
        self._register_void("mega")
        self._register_value_argless("get_false")
        self._register_void("set_is_happy")
        self._register_void("mega_f32")
        self._register_void("mega_i32")
        self._register_void("draw")
        self._register_void_argless("blocked_alrm")
        self._register_void("spawn")
        self._register_value("has_resource")
        self._register_value("has_entity")
        self._register_value("has_string")
        self._register_value_argless("get_opponent")
        self._register_void("set_d")
        self._register_void("set_opponent")
        self._register_void("motherload")
        self._register_void("motherload_subless")
        self._register_void("offset_32_bit_f32")
        self._register_void("offset_32_bit_i32")
        self._register_void("offset_32_bit_string")
        self._register_void("talk")
        self._register_value("get_position")
        self._register_void("set_position")
        self._register_void_argless("cause_game_fn_error")
        self._register_void_argless("call_on_b_fn")
        self._register_void("store")
        self._register_value_argless("retrieve")
        self._register_value("box_number")

    def _get_c_fn(self, name: str):
        return self.grug_lib["game_fn_" + name]

    def _get_c_args(self, *args: GrugValue):
        c_args = (GrugValueUnion * len(args))()
        keepalive: List[bytes] = []

        for i, v in enumerate(args):
            if isinstance(v, float):
                c_args[i]._number = v
            elif isinstance(v, bool):
                c_args[i]._bool = v
            elif isinstance(v, str):
                b = v.encode()
                keepalive.append(b)
                c_args[i]._string = ctypes.c_char_p(b)
            else:
                assert isinstance(v, int)
                c_args[i]._id = ctypes.c_uint64(v)

        return c_args, keepalive

    def _unpack_workaround(
        self, c_workaround: GrugValueWorkaround, return_type: str
    ) -> GrugValue:
        """
        Creates a GrugValueUnion, and copies the bits from GrugValueWorkaround into it.
        See the GrugValueWorkaround class docs for more information.
        """
        value = GrugValueUnion()
        ctypes.memmove(
            ctypes.byref(value), ctypes.byref(c_workaround), ctypes.sizeof(value)
        )
        return c_to_py_value(value, return_type)

    def _get_return_type(self, name: str):
        return self.state.mod_api["game_functions"][name].get("return_type")

    def _register_void(self, name: str):
        c_fn = self._get_c_fn(name)

        c_fn.argtypes = (ctypes.POINTER(GrugValueUnion),)
        c_fn.restype = None

        def fn(*args: GrugValue):
            c_args, _keepalive = self._get_c_args(*args)
            c_fn(c_args)
            if _grug_runtime_err is not None:
                raise _grug_runtime_err

        self.state._register_game_fn(name, fn)  # pyright: ignore[reportPrivateUsage]

    def _register_void_argless(self, name: str):
        c_fn = self._get_c_fn(name)

        c_fn.argtypes = ()
        c_fn.restype = None

        def fn(*args: GrugValue):
            c_fn()
            if _grug_runtime_err is not None:
                raise _grug_runtime_err

        self.state._register_game_fn(name, fn)  # pyright: ignore[reportPrivateUsage]

    def _register_value(self, name: str):
        c_fn = self._get_c_fn(name)

        c_fn.argtypes = (ctypes.POINTER(GrugValueUnion),)
        c_fn.restype = GrugValueWorkaround

        return_type = self._get_return_type(name)

        def fn(*args: GrugValue):
            c_args, _keepalive = self._get_c_args(*args)
            result: GrugValueWorkaround = c_fn(*c_args)
            if _grug_runtime_err is not None:
                raise _grug_runtime_err
            return self._unpack_workaround(result, return_type)

        self.state._register_game_fn(name, fn)  # pyright: ignore[reportPrivateUsage]

    def _register_value_argless(self, name: str):
        c_fn = self._get_c_fn(name)

        c_fn.argtypes = ()
        c_fn.restype = GrugValueWorkaround

        return_type = self._get_return_type(name)

        def fn(*args: GrugValue):
            result: GrugValueWorkaround = c_fn()
            if _grug_runtime_err is not None:
                raise _grug_runtime_err
            return self._unpack_workaround(result, return_type)

        self.state._register_game_fn(name, fn)  # pyright: ignore[reportPrivateUsage]


# Enables stepping through code with VS Code its Python debugger.
if __name__ == "__main__":
    pytest.main(sys.argv)



import json
from io import StringIO
from typing import Any, Dict, List, Union

from .frontend import (
    Argument,
    Ast,
    BinaryExpr,
    BreakStatement,
    CallExpr,
    CallStatement,
    CommentStatement,
    ContinueStatement,
    EmptyLineStatement,
    EntityExpr,
    Expr,
    FalseExpr,
    HelperFn,
    IdentifierExpr,
    IfStatement,
    LogicalExpr,
    NumberExpr,
    OnFn,
    ParenthesizedExpr,
    ResourceExpr,
    ReturnStatement,
    Statement,
    StringExpr,
    TrueExpr,
    UnaryExpr,
    VariableStatement,
    WhileStatement,
)


class Serializer:
    """Serializes AST to JSON text or grug source code."""

    @staticmethod
    def _serialize_expr(expr: Expr) -> Dict[str, Any]:
        """Serialize an expression node to dict."""
        result: Dict[str, Any] = {}

        # Handle different expression types
        if isinstance(expr, TrueExpr):
            result["type"] = "TRUE_EXPR"
        elif isinstance(expr, FalseExpr):
            result["type"] = "FALSE_EXPR"
        elif isinstance(
            expr,
            (
                StringExpr,
                ResourceExpr,
                EntityExpr,
                IdentifierExpr,
            ),
        ):
            if isinstance(expr, StringExpr):
                result["type"] = "STRING_EXPR"
            elif isinstance(expr, ResourceExpr):
                result["type"] = "RESOURCE_EXPR"
            elif isinstance(expr, EntityExpr):
                result["type"] = "ENTITY_EXPR"
            else:
                assert isinstance(expr, IdentifierExpr)
                result["type"] = "IDENTIFIER_EXPR"

            result["str"] = (
                expr.name if isinstance(expr, IdentifierExpr) else expr.string
            )
        elif isinstance(expr, NumberExpr):
            result["type"] = "NUMBER_EXPR"
            result["value"] = expr.string
        elif isinstance(expr, UnaryExpr):
            result["type"] = "UNARY_EXPR"
            result["operator"] = expr.operator.name
            result["expr"] = Serializer._serialize_expr(expr.expr)
        elif isinstance(expr, (BinaryExpr, LogicalExpr)):
            result["type"] = (
                "BINARY_EXPR" if isinstance(expr, BinaryExpr) else "LOGICAL_EXPR"
            )
            result["left_expr"] = Serializer._serialize_expr(expr.left_expr)
            result["operator"] = expr.operator.name
            result["right_expr"] = Serializer._serialize_expr(expr.right_expr)
        elif isinstance(expr, CallExpr):
            result["type"] = "CALL_EXPR"
            result["name"] = expr.fn_name
            if expr.arguments:
                result["arguments"] = [
                    Serializer._serialize_expr(arg) for arg in expr.arguments
                ]
        else:
            assert isinstance(expr, ParenthesizedExpr)
            result["type"] = "PARENTHESIZED_EXPR"
            result["expr"] = Serializer._serialize_expr(expr.expr)

        return result

    @staticmethod
    def _serialize_statement(stmt: Statement) -> Dict[str, Any]:
        """Serialize a statement node to dict."""
        result: Dict[str, Any] = {}

        if isinstance(stmt, VariableStatement):
            result["type"] = "VARIABLE_STATEMENT"
            result["name"] = stmt.name
            if stmt.type:
                result["variable_type"] = stmt.type_name
            result["assignment"] = Serializer._serialize_expr(stmt.expr)
        elif isinstance(stmt, CallStatement):
            result["type"] = "CALL_STATEMENT"
            call_expr = stmt.expr
            result["name"] = call_expr.fn_name
            if call_expr.arguments:
                result["arguments"] = [
                    Serializer._serialize_expr(arg) for arg in call_expr.arguments
                ]
        elif isinstance(stmt, IfStatement):
            result["type"] = "IF_STATEMENT"
            result["condition"] = Serializer._serialize_expr(stmt.condition)
            if stmt.if_body:
                result["if_statements"] = [
                    Serializer._serialize_statement(s) for s in stmt.if_body
                ]
            if stmt.else_body:
                result["else_statements"] = [
                    Serializer._serialize_statement(s) for s in stmt.else_body
                ]
        elif isinstance(stmt, ReturnStatement):
            result["type"] = "RETURN_STATEMENT"
            if stmt.value:
                result["expr"] = Serializer._serialize_expr(stmt.value)
        elif isinstance(stmt, WhileStatement):
            result["type"] = "WHILE_STATEMENT"
            result["condition"] = Serializer._serialize_expr(stmt.condition)
            result["statements"] = [
                Serializer._serialize_statement(s) for s in stmt.body_statements
            ]
        elif isinstance(stmt, CommentStatement):
            result["type"] = "COMMENT_STATEMENT"
            result["comment"] = stmt.string
        elif isinstance(stmt, BreakStatement):
            result["type"] = "BREAK_STATEMENT"
        elif isinstance(stmt, ContinueStatement):
            result["type"] = "CONTINUE_STATEMENT"
        else:
            assert isinstance(stmt, EmptyLineStatement)
            result["type"] = "EMPTY_LINE_STATEMENT"

        return result

    @staticmethod
    def _serialize_arguments(arguments: List[Argument]) -> List[Dict[str, str]]:
        """Serialize function arguments to list of dicts."""
        return [{"name": arg.name, "type": arg.type_name} for arg in arguments]

    @staticmethod
    def _serialize_global_statement(
        global_stmt: Union[
            VariableStatement, EmptyLineStatement, CommentStatement, OnFn, HelperFn
        ],
    ) -> Dict[str, Any]:
        """Serialize a global statement node to dict."""
        result: Dict[str, Any] = {}

        if isinstance(global_stmt, OnFn):
            result["type"] = "GLOBAL_ON_FN"
            result["name"] = global_stmt.fn_name
            if global_stmt.arguments:
                result["arguments"] = Serializer._serialize_arguments(
                    global_stmt.arguments
                )
            result["statements"] = [
                Serializer._serialize_statement(s) for s in global_stmt.body_statements
            ]
        elif isinstance(global_stmt, HelperFn):
            result["type"] = "GLOBAL_HELPER_FN"
            result["name"] = global_stmt.fn_name
            if global_stmt.arguments:
                result["arguments"] = Serializer._serialize_arguments(
                    global_stmt.arguments
                )
            if global_stmt.return_type:
                result["return_type"] = global_stmt.return_type_name
            result["statements"] = [
                Serializer._serialize_statement(s) for s in global_stmt.body_statements
            ]
        elif isinstance(global_stmt, VariableStatement):
            result["type"] = "GLOBAL_VARIABLE"
            result["name"] = global_stmt.name
            result["variable_type"] = global_stmt.type_name
            result["assignment"] = Serializer._serialize_expr(global_stmt.expr)
        elif isinstance(global_stmt, CommentStatement):
            result["type"] = "GLOBAL_COMMENT"
            result["comment"] = global_stmt.string
        else:
            assert isinstance(global_stmt, EmptyLineStatement)
            result["type"] = "GLOBAL_EMPTY_LINE"

        return result

    @staticmethod
    def ast_to_json_text(ast: Ast) -> str:
        """Convert AST to JSON text representation."""
        serialized = [Serializer._serialize_global_statement(node) for node in ast]
        return json.dumps(serialized, separators=(",", ":"))
        # return json.dumps(serialized, indent=4)  # TODO: Ensure this passes all grug-tests tests too!

    @staticmethod
    def ast_to_grug(ast: List[Dict[str, Any]]) -> str:
        """Convert AST to grug source code."""
        output = StringIO()
        indentation = [0]  # Using list to make it mutable in nested functions

        def write(text: str) -> None:
            """Write text to output."""
            output.write(text)

        def apply_indentation() -> None:
            """Apply current indentation level."""
            write("    " * indentation[0])

        def apply_expr(expr: Dict[str, Any]) -> None:
            """Generate code for an expression."""
            assert isinstance(expr, dict), "Expression must be a dict"
            assert "type" in expr, "Expression must have a 'type' field"

            expr_type = expr["type"]

            if expr_type == "TRUE_EXPR":
                write("true")
            elif expr_type == "FALSE_EXPR":
                write("false")
            elif expr_type == "STRING_EXPR":
                write(f'"{expr["str"]}"')
            elif expr_type == "IDENTIFIER_EXPR":
                write(expr["str"])
            elif expr_type == "NUMBER_EXPR":
                write(expr["value"])
            elif expr_type == "UNARY_EXPR":
                operator = expr["operator"]
                if operator == "MINUS_TOKEN":
                    write("-")
                elif operator == "NOT_TOKEN":
                    write("not ")
                apply_expr(expr["expr"])
            elif expr_type == "BINARY_EXPR":
                apply_expr(expr["left_expr"])
                operator = get_binary_operator(expr["operator"])
                write(f" {operator} ")
                apply_expr(expr["right_expr"])
            elif expr_type == "LOGICAL_EXPR":
                apply_expr(expr["left_expr"])
                operator = get_logical_operator(expr["operator"])
                write(f" {operator} ")
                apply_expr(expr["right_expr"])
            elif expr_type == "CALL_EXPR":
                write(f'{expr["name"]}(')
                if "arguments" in expr:
                    for i, arg in enumerate(expr["arguments"]):
                        if i > 0:
                            write(", ")
                        apply_expr(arg)
                write(")")
            elif expr_type == "PARENTHESIZED_EXPR":
                write("(")
                apply_expr(expr["expr"])
                write(")")

        def get_logical_operator(token: str) -> str:
            """Convert token to logical operator."""
            if token == "AND_TOKEN":
                return "and"
            elif token == "OR_TOKEN":
                return "or"
            raise ValueError(f"Unknown logical operator: {token}")

        def get_binary_operator(token: str) -> str:
            """Convert token to binary operator."""
            operators = {
                "PLUS_TOKEN": "+",
                "MINUS_TOKEN": "-",
                "MULTIPLICATION_TOKEN": "*",
                "DIVISION_TOKEN": "/",
                "EQUALS_TOKEN": "==",
                "NOT_EQUALS_TOKEN": "!=",
                "GREATER_OR_EQUAL_TOKEN": ">=",
                "GREATER_TOKEN": ">",
                "LESS_OR_EQUAL_TOKEN": "<=",
                "LESS_TOKEN": "<",
            }
            if token in operators:
                return operators[token]
            raise ValueError(f"Unknown binary operator: {token}")

        def try_get_else_if(
            else_statements: List[Dict[str, Any]],
        ) -> Union[Dict[str, Any], None]:
            """Check if else block is actually an else-if."""
            if (
                len(else_statements) > 0
                and else_statements[0].get("type") == "IF_STATEMENT"
            ):
                return else_statements[0]
            return None

        def apply_comment(statement: Dict[str, Any]) -> None:
            """Generate code for a comment."""
            write(f'# {statement["comment"]}\n')

        def apply_if_statement(statement: Dict[str, Any]) -> None:
            """Generate code for an if statement."""
            write("if ")
            apply_expr(statement["condition"])
            write(" {\n")

            if "if_statements" in statement:
                apply_statements(statement["if_statements"])

            if "else_statements" in statement:
                apply_indentation()
                write("} else ")

                else_if_node = try_get_else_if(statement["else_statements"])
                if else_if_node:
                    apply_if_statement(else_if_node)
                else:
                    write("{\n")
                    apply_statements(statement["else_statements"])
                    apply_indentation()
                    write("}\n")
            else:
                apply_indentation()
                write("}\n")

        def apply_statement(statement: Dict[str, Any]) -> None:
            """Generate code for a statement."""
            stmt_type = statement["type"]

            if stmt_type == "VARIABLE_STATEMENT":
                write(statement["name"])

                if "variable_type" in statement:
                    write(f': {statement["variable_type"]}')

                if "assignment" in statement:
                    write(" = ")
                    apply_expr(statement["assignment"])

                write("\n")

            elif stmt_type == "CALL_STATEMENT":
                write(f'{statement["name"]}(')
                if "arguments" in statement:
                    for i, arg in enumerate(statement["arguments"]):
                        if i > 0:
                            write(", ")
                        apply_expr(arg)
                write(")\n")

            elif stmt_type == "IF_STATEMENT":
                apply_if_statement(statement)

            elif stmt_type == "RETURN_STATEMENT":
                write("return")
                if "expr" in statement:
                    write(" ")
                    apply_expr(statement["expr"])
                write("\n")

            elif stmt_type == "WHILE_STATEMENT":
                write("while ")
                apply_expr(statement["condition"])
                write(" {\n")
                apply_statements(statement["statements"])
                apply_indentation()
                write("}\n")

            elif stmt_type == "BREAK_STATEMENT":
                write("break\n")

            elif stmt_type == "CONTINUE_STATEMENT":
                write("continue\n")

            elif stmt_type == "COMMENT_STATEMENT":
                apply_comment(statement)

        def apply_statements(statements: List[Dict[str, Any]]) -> None:
            """Generate code for a list of statements."""
            indentation[0] += 1

            for statement in statements:
                if statement["type"] == "EMPTY_LINE_STATEMENT":
                    write("\n")
                else:
                    apply_indentation()
                    apply_statement(statement)

            indentation[0] -= 1

        def apply_arguments(arguments: List[Dict[str, Any]]) -> None:
            """Generate code for function arguments."""
            for i, arg in enumerate(arguments):
                if i > 0:
                    write(", ")
                write(f'{arg["name"]}: {arg["type"]}')

        def apply_helper_fn(statement: Dict[str, Any]) -> None:
            """Generate code for a helper function."""
            write(f'{statement["name"]}(')

            if "arguments" in statement:
                apply_arguments(statement["arguments"])

            write(")")

            if "return_type" in statement:
                write(f' {statement["return_type"]}')

            write(" {\n")

            if "statements" in statement:
                apply_statements(statement["statements"])

            write("}\n")

        def apply_on_fn(statement: Dict[str, Any]) -> None:
            """Generate code for an on function."""
            write(f'{statement["name"]}(')

            if "arguments" in statement:
                apply_arguments(statement["arguments"])

            write(") {\n")

            if "statements" in statement:
                apply_statements(statement["statements"])

            write("}\n")

        def apply_global_variable(statement: Dict[str, Any]) -> None:
            """Generate code for a global variable."""
            write(f'{statement["name"]}: {statement["variable_type"]} = ')
            apply_expr(statement["assignment"])
            write("\n")

        def apply_root(root: List[Dict[str, Any]]) -> None:
            """Generate code for the root AST node."""
            for statement in root:
                stmt_type = statement["type"]

                if stmt_type == "GLOBAL_VARIABLE":
                    apply_global_variable(statement)
                elif stmt_type == "GLOBAL_ON_FN":
                    apply_on_fn(statement)
                elif stmt_type == "GLOBAL_HELPER_FN":
                    apply_helper_fn(statement)
                elif stmt_type == "GLOBAL_EMPTY_LINE":
                    write("\n")
                elif stmt_type == "GLOBAL_COMMENT":
                    apply_comment(statement)

        # Main execution
        apply_root(ast)
        return output.getvalue()



import json
import sys
from enum import Enum, auto
from pathlib import Path
from typing import Any, Callable, Dict, Sequence, cast

from grug.game_fn import GameFn
from grug.grug_dir import GrugDir
from grug.grug_file import GrugFile
from grug.grug_package import GrugPackage

from .frontend import Frontend, HelperFn, OnFn, Parser, Tokenizer, VariableStatement
from .serializer import Serializer


class GrugRuntimeErrorType(Enum):
    STACK_OVERFLOW = 0  # Using auto() here would assign 1
    TIME_LIMIT_EXCEEDED = auto()
    GAME_FN_ERROR = auto()


GrugRuntimeErrorHandler = Callable[[str, GrugRuntimeErrorType, str, str], None]


def default_runtime_error_handler(
    reason: str,
    grug_runtime_error_type: GrugRuntimeErrorType,
    on_fn_name: str,
    on_fn_path: str,
):
    print(
        f"grug runtime error in {on_fn_name}(): {reason}, in {on_fn_path}",
        file=sys.stderr,
    )


class GrugState:
    def __init__(
        self,
        *,
        runtime_error_handler: GrugRuntimeErrorHandler,
        mod_api_path: str,
        mods_dir_path: str,
        on_fn_time_limit_ms: float,
        packages: Sequence[GrugPackage],
    ):
        self.runtime_error_handler = runtime_error_handler

        with open(mod_api_path) as f:
            raw = json.load(f)
        if not isinstance(raw, dict):
            exit("Error: mod API JSON root must be an object")
        self.mod_api: Dict[str, Any] = cast(Dict[str, Any], raw)

        self._assert_mod_api()

        self.mods_dir_path = mods_dir_path

        self.on_fn_time_limit_ms = on_fn_time_limit_ms

        self.frontend = Frontend(self.mod_api)

        self.game_fns: Dict[str, GameFn] = {}
        self._add_game_fns_from_packages(packages)

        self.next_id = 0

        self.fn_depth = 0

    def _assert_mod_api(self):
        entities = self.mod_api.get("entities")
        if not isinstance(entities, dict):
            exit("Error: 'entities' must be a JSON object")

        entities_dict = cast(Dict[str, Any], entities)
        self._assert_entities_sorted(entities_dict)

        for entity_name, entity in entities_dict.items():
            if not isinstance(entity, dict):
                exit(f"Error: entity '{entity_name}' must be a JSON object")

            entity_dict = cast(Dict[str, Any], entity)
            on_functions = entity_dict.get("on_functions")
            if on_functions is None:
                continue

            if not isinstance(on_functions, dict):
                exit(
                    f"Error: 'on_functions' for entity '{entity_name}' must be a JSON object"
                )

            on_functions_dict = cast(Dict[str, Any], on_functions)
            self._assert_on_functions_sorted(entity_name, on_functions_dict)

        game_functions = self.mod_api.get("game_functions")
        if not isinstance(game_functions, dict):
            exit("Error: 'game_functions' must be a JSON object")

        game_functions_dict = cast(Dict[str, Any], game_functions)
        self._assert_game_functions_sorted(game_functions_dict)

    def _assert_entities_sorted(self, entities: Dict[str, Any]):
        keys = list(entities.keys())
        sorted_keys = sorted(keys)

        if keys != sorted_keys:
            for actual, expected in zip(keys, sorted_keys):
                if actual != expected:
                    exit(
                        f"Error: Entities must be sorted alphabetically in mod_api.json, "
                        f"so '{expected}' must come before '{actual}'"
                    )
            assert False

    def _assert_on_functions_sorted(
        self, entity_name: str, on_functions: Dict[str, Any]
    ):
        keys = list(on_functions.keys())
        sorted_keys = sorted(keys)

        if keys != sorted_keys:
            for actual, expected in zip(keys, sorted_keys):
                if actual != expected:
                    exit(
                        "Error: on_functions for entity "
                        f"'{entity_name}' must be sorted alphabetically in mod_api.json, "
                        f"so '{expected}' must come before '{actual}'"
                    )
            assert False

    def _assert_game_functions_sorted(self, game_functions: Dict[str, Any]):
        keys = list(game_functions.keys())
        sorted_keys = sorted(keys)

        if keys != sorted_keys:
            for actual, expected in zip(keys, sorted_keys):
                if actual != expected:
                    exit(
                        f"Error: Game functions must be sorted alphabetically in mod_api.json, "
                        f"so {expected}() must come before {actual}()"
                    )
            assert False

    def _add_game_fns_from_packages(self, packages: Sequence[GrugPackage]):
        for pkg in packages:
            for game_fn in pkg.game_fns:
                if game_fn.__name__ in self.game_fns:
                    exit(
                        f"Error: Game function '{game_fn.__name__}' has already been registered, so you either registered it twice, or its grug package prefix clashes with another grug package"
                    )

                name = (
                    f"{pkg.prefix}_{game_fn.__name__}"
                    if pkg.prefix
                    else game_fn.__name__
                )
                self._register_game_fn(name, game_fn)

    def game_fn(self, fn: GameFn) -> GameFn:
        """Decorator for game functions."""
        self._register_game_fn(fn.__name__, fn)
        return fn

    def _register_game_fn(self, name: str, fn: GameFn):
        self.game_fns[name] = fn

    def compile_grug_file(self, grug_file_relative_path: str):
        """Read a file and pass its contents to the frontend."""
        mod = Path(grug_file_relative_path).parts[0]

        grug_file_absolute_path = Path(self.mods_dir_path) / grug_file_relative_path
        text = grug_file_absolute_path.read_text()

        entity_type = self._get_file_entity_type(Path(grug_file_relative_path).name)

        ast = self.frontend.compile_grug_file(text, mod, entity_type)

        global_variables = [s for s in ast if isinstance(s, VariableStatement)]

        on_fns = {s.fn_name: s for s in ast if isinstance(s, OnFn)}

        helper_fns = {s.fn_name: s for s in ast if isinstance(s, HelperFn)}

        game_fn_return_types = {
            fn_name: fn.get("return_type")
            for fn_name, fn in self.mod_api["game_functions"].items()
        }

        return GrugFile(
            grug_file_relative_path,
            mod,
            global_variables,
            on_fns,
            helper_fns,
            self.game_fns,
            game_fn_return_types,
            self,
        )

    def _get_file_entity_type(self, grug_filename: str) -> str:
        """
        Extract and validate the entity type from a grug filename.

        Args:
            grug_filename: A filename like 'furnace-BlockEntity.grug'

        Returns:
            The entity type string (e.g., 'BlockEntity')

        Raises:
            ValueError: If the filename format is invalid
        """
        # Find the dash
        dash_index = grug_filename.find("-")

        if dash_index == -1 or dash_index + 1 >= len(grug_filename):
            raise ValueError(
                f"'{grug_filename}' is missing an entity type in its name; "
                f"use a dash to specify it, like 'ak47-gun.grug'"
            )

        # Find the period after the dash
        period_index = grug_filename.find(".", dash_index + 1)

        if period_index == -1:
            raise ValueError(f"'{grug_filename}' is missing a period in its filename")

        # Extract entity type (between dash and period)
        entity_type = grug_filename[dash_index + 1 : period_index]

        # Check if entity type is empty
        if len(entity_type) == 0:
            raise ValueError(
                f"'{grug_filename}' is missing an entity type in its name; "
                f"use a dash to specify it, like 'ak47-gun.grug'"
            )

        # Validate PascalCase
        self._check_custom_id_is_pascal(entity_type)

        return entity_type

    def _check_custom_id_is_pascal(self, type_name: str):
        """
        Validate that a custom ID type name is in PascalCase.

        Args:
            type_name: The type name to validate

        Raises:
            ValueError: If the type name is not valid PascalCase
        """
        # The first character must always be uppercase
        if not type_name[0].isupper():
            raise ValueError(
                f"'{type_name}' seems like a custom ID type, but isn't in PascalCase"
            )

        # Custom IDs only consist of uppercase, lowercase characters, and digits
        for c in type_name:
            if not (c.isupper() or c.islower() or c.isdigit()):
                raise ValueError(
                    f"'{type_name}' seems like a custom ID type, but it contains '{c}', "
                    f"which isn't uppercase/lowercase/a digit"
                )

    def compile_all_mods(self) -> GrugDir:
        """
        Compiles all grug mods under self.mods_dir_path recursively.

        Returns:
            GrugDir: Root directory representing the entire mods/ folder.
        """
        mods_path = Path(self.mods_dir_path)

        def compile_dir(current_path: Path, dir_name: str) -> GrugDir:
            grug_dir = GrugDir(name=dir_name)

            for entry in current_path.iterdir():
                if entry.is_dir():
                    subdir = compile_dir(entry, entry.name)
                    grug_dir.dirs[entry.name] = subdir
                elif entry.is_file() and entry.suffix == ".grug":
                    relative_path = entry.relative_to(mods_path).as_posix()
                    grug_file = self.compile_grug_file(relative_path)
                    grug_dir.files[relative_path] = grug_file

            return grug_dir

        root_dir = GrugDir(name="mods")
        for mod_dir in mods_path.iterdir():
            if mod_dir.is_dir():
                root_dir.dirs[mod_dir.name] = compile_dir(mod_dir, mod_dir.name)

        return root_dir

    def update(self):
        # TODO: Implement hot reloading
        pass

    def run_all_package_tests(self):
        mods = self.compile_all_mods()

        tests_ran = 0

        def run(dir: GrugDir):
            for subdir in dir.dirs.values():
                run(subdir)
            for file in dir.files.values():
                print(f"Testing {file.relative_path}...")
                test = file.create_entity()
                test.on_run()
                nonlocal tests_ran
                tests_ran += 1

        run(mods)

        print(f"All {tests_ran} tests passed!")

    # TODO: Should this method be moved out of this GrugState, so it becomes a free function?
    def dump_file_to_json(self, input_grug_path: str, output_json_path: str):
        grug_text = Path(input_grug_path).read_text()

        tokens = Tokenizer(grug_text).tokenize()

        ast = Parser(tokens).parse()

        json_text = Serializer.ast_to_json_text(ast)

        Path(output_json_path).write_text(json_text)

        return False

    # TODO: Should this method be moved out of this GrugState, so it becomes a free function?
    def generate_file_from_json(self, input_json_path: str, output_grug_path: str):
        json_text = Path(input_json_path).read_text()

        ast = json.loads(json_text)

        grug_text = Serializer.ast_to_grug(ast)

        Path(output_grug_path).write_text(grug_text)

        return False



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict

from grug.grug_file import GrugFile


@dataclass
class GrugDir:
    """Represents a directory of grug files and subdirectories."""

    name: str
    files: Dict[str, GrugFile] = field(default_factory=lambda: {})
    dirs: Dict[str, GrugDir] = field(default_factory=lambda: {})
